# -*- tcl -*- 
#
# This code decodes the windows .lnk file.  It provides several bits of
# information about the link as defined below.
#
# Return value is 1 if "filename" is a lnk file. 
# Otherwise 0 (includes unreadable not a filename, etc.)
# If 0 is returned and "to" is provided, "to" will be set to a reason.
# "to" if provided is a var that will be set to the file the link points to.
# "dir" if provided will be set to 1 if the link points at a directory, else 0.

proc setTo {to this} {
  if {$to != {}} {
    uplevel 2 "set $to \"$this\""
  }
}

proc LnkFile { filename {to {}} {dir {}} {try 0} {diag 0}} {

  # Constants of interest...
  # set headerlength 72  # 0x4c
  # set clid0 [expr {0x21401}]
  # set clid1 0
  # set clid2 [expr { 0xc0}]
  # set clid3 [expr { 0x46000000 }]
  # linkflags bits ...
   set HasLinkTargetIDList 1
   set HasLinkInfo         2
   set Name                4
   set RelativePath        8
   set WorkingDir         16
   set Arguments          32
   set IconLocation       64
   set IsUniCode         128
  # All we are intested in here...
  # FileAttributesFlags ....
  # set ReadOnly            1
  # set Hidden              2
  # set System              4
   set Directory          16  
  # set Archive            32  # 0x20
  # set Normal            128  # 0x80

  # first we want to be sure that it really is a .lnk file...
  # we don't trust, or even test, the suffix, but rather use the
  # internal info in the file.
  if { [file exists $filename] == 0 || [file readable $filename] == 0 } {
    frputs "LnkFile no such file or not readable  " filename
    setTo $to "no such file ($filename) or not readable"
    return 0
  }
  set f [open $filename]
  chan configure $f -translation binary
#  chan seek $f
  set len [BinaryRead4 $f]
  if { $len != 0x4c && ! $try} {
    frputs "LnkFile Not link file, header length $len, not 76  " filename
    setTo %to "Not link file, header length $len, not 76"
    return 0
  }
  set clid0 [BinaryRead4 $f]
  set clid1 [BinaryRead4 $f]
  set clid2 [BinaryRead4 $f]
  set clid3 [BinaryRead4 $f]
  if { ($clid0 != 0x21401 || $clid1 != 0 || \
	   $clid2 != 0xc0 || $clid3 != 0x46000000) && ! $try } {
    frputs "LnkFile Not a link file, CLID is bad  " filename
    setTo $to "Not a link file, CLID is bad"
    return 0  
  }

  set linkflags [BinaryRead4 $f]
  set attrib  [BinaryRead4 $f]
  # We are not interested in the rest of the header, seek to the next
  # block...
   PutsHex linkflags
   PutsHex attrib
   set IsDir [expr {($attrib & $Directory) != 0?1:0}]
   setTo $dir $IsDir

   # Life should be so easy, push onward...

   chan seek $f 76 start

   # Here we find optional fields in this order 
   # LinkTarget_IDList, LinkInfo, StringData

   # LinkTarget_IDList...
  if { $linkflags & $HasLinkTargetIDList } {
     # we just skip this....
     set Start_of_TargetID [chan tell $f]
     set ListSize  [BinaryRead2 $f]
     PutsHex Start_of_TargetID
     PutsHex ListSize
     chan seek $f  [expr {$ListSize } ] current

     #    set ListInfo [chan read $f [expr {$ListSize-2 } ]
     #       puts "LinkTarget_IDList $ListSize >$ListInfo<"

     # Here is an array of IDList things with the first 2-bytes indicating the
     # length which includes the length itself (zero is the terminal)

     set End_of_TargetID [chan tell $f]
     PutsHex End_of_TargetID
     #    set IDindex 0
     #   while { [set IDsize [BinaryRead2 $f]] > 0 } {
     #      set IDarray([incr IDindex]) [chan read $f [incr $IDsize -2]]
     #    }
     #   puts "Found $IDindex structures skipped"
     #    while { $IDindex > 0 } {
     #      puts "$IDarray($IDindex)"
     #      incr IDindex -1
     #    }
   }
   # LinkInfo  Finally getting to the meat of it...
   if { $linkflags & $HasLinkInfo } {
     set LinkInfoOffset [chan tell $f]
     set LinkInfoSize [BinaryRead4 $f]
     set LinkInfoHeaderSize [BinaryRead4 $f]
     set LinkInfoFlags [BinaryRead4 $f]
     set LinkInfoVolIDOffset [BinaryRead4 $f]
   
     set LinkInfoLocalBasePathOffset [BinaryRead4 $f]
     set LinkInfoCommonNetworkOffset [BinaryRead4 $f]
     set LinkInfoCommonPathSuffixOffset [BinaryRead4 $f]
     if { $LinkInfoHeaderSize != 0x1c } {
       set LinkInfoLocalBasePathOffsetUcd [BinaryRead4 $f]
       set LinkInfoCommonPathSuffixOffsetUcd [BinaryRead4 $f]
     } else {
       set LinkInfoLocalBasePathOffsetUcd 0
       set LinkInfoCommonPathSuffixOffsetUcd 0
     }
     #     puts "$linkflags $HasLinkInfo "
     PutsHex LinkInfoOffset
     PutsHex LinkInfoSize
     PutsHex LinkInfoHeaderSize
     PutsHex LinkInfoFlags
     PutsHex LinkInfoVolIDOffset
     PutsHex LinkInfoLocalBasePathOffset
     PutsHex LinkInfoCommonNetworkOffset
     PutsHex LinkInfoCommonPathSuffixOffset
     PutsHex LinkInfoLocalBasePathOffsetUcd
     PutsHex LinkInfoCommonPathSuffixOffsetUcd

     # Ok, lets get the Volume info... if it exists
     if { $LinkInfoVolIDOffset != 0 } {
       chan seek $f [set VolIdOffset \
			 [expr {$LinkInfoVolIDOffset + $LinkInfoOffset} ]]
       PutsHex VolIdOffset
       set VolIDSize [BinaryRead4 $f]
       set DriveType  [BinaryRead4 $f]
       set DriveType [switch  $DriveType { 
	 0 {expr {{UnKnown}}}
	 1 {expr {{No_Root_Dir}}}
	 2 {expr {{Removable}}}
	 3 {expr {{Fixed}}}
	 4 {expr {{Remote}}}
	 5 {expr {{CDROM}}}
	 6 {expr {{RamDisk}}}
       } ]
       PutsString DriveType
       set DriveSerial [BinaryRead4 $f]
       PutsHex DriveSerial
       set VolLabelOffset  [BinaryRead4 $f]
       if { $VolLabelOffset == 0x14 } {
	 
	      set VolLabelOffset  [BinaryRead4 $f]

	      set VolIsUcode 1
	    } else {
	      set VolIsUcode 0
	    }
       PutsHex VolLabelOffset
       PutsHex VolIsUcode
       if { $VolIsUcode == 0 } {
	      set VolIdString [GetNullTermString $f \
				   [expr { $VolIdOffset + $VolLabelOffset } ]]
       } else {
	      set VolIdString [GetNullTermStringUnc $f \
				   [expr { $VolIdOffset + $VolLabelOffset } ]]
       }
       PutsString VolIdString
     }
     # Here we get the Local Base string
     if { $LinkInfoLocalBasePathOffset != 0 } {
       set BasePath [GetNullTermString $f \
			 [expr {$LinkInfoLocalBasePathOffset + \
				    $LinkInfoOffset } ]]
       PutsString BasePath
     }
     # Now the common Network info...
     if { $LinkInfoCommonNetworkOffset != 0 } {
       set NetOffset [chan tell $f]
       set NetLinkSize  [BinaryRead4 $f]
       set NetLinkFlags [BinaryRead4 $f]
       set NetNameOffset [BinaryRead4 $f]
       set DevNameOffset [BinaryRead4 $f]
       set NetProviderType [BinaryRead4 $f]
       PutsHex NetLinkSize
       PutsHex NetOffset
       PutsHex NetLinkFlags 
       PutsHex NetNameOffset
       PutsHex DevNameOffset
       PutsHex NetProviderType
       if { $NetNameOffset > 0x14 } {
	      set NetNameOffsetUnc [BinaryRead4 $f]
	      set DevNameOffsetUn [BinaryRead4 $f]
       } else {
	      set NetNameOffsetUnc 0
	      set DevNameOffsetUn 0
       }
       PutsHex NetNameOffsetUnc
       PutsHex DevNameOffsetUn
     
       set NetName [GetNullTermString $f \
			[expr { $NetOffset + $NetNameOffset}]]
       PutsString NetName
       set DevName  [GetNullTermString $f \
			 [expr { $NetOffset + $DevNameOffset}]]
       if {$DevName == "\"" } {set DevName {}}
       PutsString DevName
       if { $NetNameOffsetUnc != 0 } {
         set NetNameUnc [GetNullTermStringUnc $f \
			  [expr { $NetOffset + $NetNameOffset}]]
         PutsString NetNameUnc
         set DevNameUnc  [GetNullTermStringUnc $f \
		 	  [expr { $NetOffset + $DevNameOffset}]]
       PutsString DevNameUnc
       }
     }
     # Common Path Suffix
     if { $LinkInfoCommonPathSuffixOffset != 0 } {
       set CommonPathSuffix \
	   [GetNullTermString $f \
		[expr {$LinkInfoOffset + $LinkInfoCommonPathSuffixOffset }]]
       PutsString CommonPathSuffix
     }
     #LinkInfoLocalBasePathOffsetUcd LinkInfoCommonPathSuffixOffsetUcd
     if { $LinkInfoLocalBasePathOffsetUcd != 0 } {
       set LocalBasePathUcd \
	   [GetNullTermStringUcd $f \
		[expr {$LinkInfoOffset + $LinkInfoLocalBasePathOffsetUcd }]]
       PutsString LocalBasePathUcd
     }
     if { $LinkInfoCommonPathSuffixOffsetUcd != 0 } {
       set CommonPathSuffixUcd \
	   [GetNullTermStringUcd $f \
		[expr {$LinkInfoOffset + $LinkInfoCommonPathSuffixOffsetUcd }]]
       PutsString CommonPathSuffixUcd
     }
   }
   # Here we look for string data.  Strings start with a count and
   # contain exactly that many characters...
   # the order is {name_string} {relative_path} {working_dir} 
   #              {command_line_arguments} {Icon_location}
   # all this is conditioned by the linkflags word...
   set val " Value"
   #  puts "doing strings $linkflags"
   foreach str {Name RelativePath WorkingDir Arguments IconLocation} {
     #	puts "[expr $$str] $str"
     if { $linkflags & [expr $$str] } {
       set len [BinaryRead2 $f]
       if { $linkflags & $IsUniCode } {
	 chan configure $f -encoding unicode
       }
       set $str$val [chan read $f $len]
       chan configure $f -encoding [encoding system]
       PutsString $str$val
     }
   }
   # So lets build a target name
   # Prefered is basePath/CommonPathSuffix
   # DevName/CommonPathSuffix
   # relativePath
   # netName/commonPathSuffix
   # if we have neither relativePath nor CommonPathSuffix we fail
   set slash /
   set result "Confused"
   if {![info exists CommonPathSuffix]} {
     if {![info exist RelativePath$val]} {
     } else {
       set result [set RelativePath$val]
     }
   } else {
     if { [info exists BasePath]} {
       set result $BasePath$CommonPathSuffix
     } elseif {[info exists DevName]} {
       set result $DevName/$CommonPathSuffix
     } elseif {[info exists NetName]} {
       set result $NetName/$CommonPathSuffix
     }				      
   }
   # Clean it up...
   #puts ">$result<"
   regsub -all {\\} $result {/} out
   regsub {/\./} $out {/} out

   # documentation of *.lnk files say we can not always trust...
   if {! $IsDir && [file exists $out] && [file isdirectory $out]} {
     set IsDir 1
   }
   #puts ">$IsDir< >$out<"
   setTo $dir $IsDir
    
   PutsHex IsDir
   setTo $to $out
   return 1 
}
proc GetNullTermString { f offset } {
  chan seek $f $offset start
  set result ""
  while { [set c [chan read $f 1]] != "\x00" } {
    set result $result$c
    if {[chan eof $f]  } {
      break
    }
  }
  return $result  
}
proc GetNullTermStringUnc { f offset } {
  chan seek $f $offset start
  set result ""
  chan configure $f -encoding unicode
  while { [set c [chan read $f 1]] != "\x00" } {
    set result $result$c
    if {[chan eof $f] } {
      break
    }
  }
  chan configure $f -encoding [encoding system]
			  
  return $result  
}


proc BinaryRead4 { f } {
  set rtn 0
  binary scan [chan read $f 4] i rtn
  return $rtn
}
proc BinaryRead2 { f } {
  set rtn 0
  binary scan [chan read $f 2] s rtn
  return $rtn
}

proc ByteToHex { char } {
  binary scan $char c c1
  return [format %2x $c1]
}
# These are debug routines enable by diag in the LnkFile call
proc PutsHex { this } {
  upvar $this here
  if {[uplevel set diag]} {
    puts "$this = [format 0x%08x $here]"
  }
}
proc PutsString { this } {
  upvar $this here
  if {[uplevel set diag]} {
    puts "$this is $here"
  }
}

#
# Here we try to verify a lnk pointer for linux
#
# we assume that the C:/ or other such ?:/ refer to the 
# mount point of (some) volume on the linux system.
# First we try the current volume. If sucessful it is entered in 
# a table for the next time.
# If that fails, we try a little table we keep to translate these.
# If that finds a file, we enter the new translation in our
# table.  If not, we fail with a null name.
#
# The other entry we decode is <??Folder> where ?? appears to be
# the current folder name.  We translate this to "."
#
# To do our work we require the mountpoint of the place the link
# was found and the found link.  We assume the link will be used
# within the current "pwd" (to make the "." work).

#
proc TranslateLnk {link curMountPt} {
  global glob
  set filePrefix {}
  switch -glob $link {
    {?:/*} {
      set first [string toupper [string range $link 0 0]]
      if {$::tcl_platform(platform) != "windows"} {
	set filePrefix $curMountPt
	set file $filePrefix/[string range $link 3 end]
      } else {
	set glob(volume,$first) [file volumes]
	set file $link
      }
    }
    {//*} {
     if {$::tcl_platform(platform) != "windows"} {
       set file $link
     } else {
       # Lets assume that it is mounted with the same name
       set file [string range $link 1 end]
     }
    }
    {*.url}   {return $link}
    {./*}     {set file $link }
    default   {set file "./$link" }
  }
  # Check if we found the file...
  if {![file exists $file]} {
    # if this is a ?:/ file and we have a previous working volume, try that
    if {$filePrefix != {} && [info exists glob(volume,$first)]} {
      foreach vol $glob(volume,$first) {     
	if {[file exists [set file $vol/[string range $link 3 end]]]} {
	  return $file
	}
      }
    }
    return {}
  } elseif {$filePrefix != {}} {
    # Allow several prefix to volume translations but only
    # one copy of a given transation.
    if {![info exists glob(volume,$first)] || \
	    $filePrefix ni $glob(volume,$first)} {
      lappend glob(volume,$first) $filePrefix
    }
  }
  return $file
}
	
